! WHIZARD 2.0.0_rc2 Sat Feb 06 2010
! 
! (C) 1999-2010 by 
!     Wolfgang Kilian <kilian@hep.physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@physik.uni-freiburg.de>
!     with contributions by Sebastian Schmidt
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_ewa

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use lorentz !NODEP!
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux

  implicit none
  private

  public :: ewa_data_t
  public :: ewa_data_init
  public :: ewa_data_write
  public :: interaction_init_ewa
  public :: interaction_apply_ewa

  integer, parameter :: NONE = 0
  integer, parameter :: ZERO_QMIN = 1
  integer, parameter :: Q_MAX_TOO_SMALL = 2
  integer, parameter :: ZERO_XMIN = 3

  type :: ewa_data_t
     private
     type(model_t), pointer :: model => null ()
     !!! Do we need this for the EWA?
     type(flavor_t) :: flv
     real(default) :: pt_max
     real(default) :: x_0
     real(default) :: x_1
     real(default) :: mass
     real(default) :: cv
     real(default) :: ca
     real(default) :: coeff
     integer :: error = 0
  end type ewa_data_t


contains

  subroutine ewa_data_init (data, model, flv, x_0, x_1, pt_max, mass)
    type(ewa_data_t), intent(inout) :: data
    type(model_t), intent(in), target :: model
    type(flavor_t), intent(in) :: flv
    real(default), intent(in) :: x_0, x_1, pt_max
    real(default), intent(in), optional :: mass
    data%model => model
    data%flv = flv
    data%pt_max = pt_max
    data%x_0 = x_0
    data%x_1 = x_1
    !!! data%cv = cv
    !!! data%ca = ca
    if (present (mass)) then
       data%mass = mass
    else
       data%mass = flavor_get_mass (flv)
    end if
    !!!
    !!! INCOMPLETE
    !!!
    !!!if (max (data%mass, data%q_min) == 0) then
    !!!   data%error = ZERO_QMIN;  return
    !!!else if (max (data%mass, data%q_min) >= data%E_max) then
    !!!   data%error = Q_MAX_TOO_SMALL;  return
    !!!end if
    !!!data%log = log (4 * (data%E_max / max (data%mass, data%q_min)) ** 2 )
    !!!data%c0 = data%alpha / pi &
    !!!     * flavor_get_charge (data%flv)**2 &
    !!!     * log (data%x_min) * (data%log - log (data%x_min))
    !!!data%c1 = data%alpha / pi &
    !!!     * flavor_get_charge (data%flv)**2 &
    !!!     * log (data%x_max) * (data%log - log (data%x_max))
  end subroutine ewa_data_init

  subroutine ewa_data_check (data)
    type(ewa_data_t), intent(in) :: data
    select case (data%error)
    case (ZERO_QMIN)
       call msg_fatal (" EWA: Particle mass is zero")
    case (Q_MAX_TOO_SMALL)
       call msg_fatal (" EWA: Particle mass exceeds Qmax")
    case (ZERO_XMIN)
       call msg_fatal (" EWA: x_min must be larger than zero")
    end select
  end subroutine ewa_data_check

  subroutine ewa_data_write (data, unit)
    type(ewa_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit);  if (u < 0)  return
    write (u, *) "EWA data:"
    write (u, *) "  prt    = ", char (flavor_get_name (data%flv))
    write (u, *) "  x_0    = ", data%x_0
    write (u, *) "  x_1    = ", data%x_1
    write (u, *) "  pt_max = ", data%pt_max
    write (u, *) "  mass   = ", data%mass
    write (u, *) "  cv     = ", data%cv
    write (u, *) "  ca     = ", data%ca
  end subroutine ewa_data_write

  subroutine interaction_init_ewa (int, data)
    type(interaction_t), intent(out) :: int
    type(ewa_data_t), dimension(:), intent(in) :: data
    type(quantum_numbers_mask_t), dimension(3) :: mask
    integer, dimension(3) :: lock
    type(polarization_t) :: pol
    type(quantum_numbers_t), dimension(1) :: qn_fc, qn_hel
    type(flavor_t) :: flv_photon
    type(quantum_numbers_t) :: qn_photon, qn
    type(state_iterator_t) :: it_hel
    integer :: i
    mask = new_quantum_numbers_mask (.false., .false., &
         mask_h = (/ .false., .false., .true. /))
    lock = (/ 2, 1, 0 /)
    call interaction_init (int, 1, 0, 2, mask=mask, lock=lock)
    call flavor_init (flv_photon, PHOTON, data(1)%model)
    call quantum_numbers_init (qn_photon, flv_photon)
    do i = 1, size (data)
       call polarization_init_generic (pol, data(i)%flv)
       call quantum_numbers_init (qn_fc(1), &
            flv = data(i)%flv, col = color_from_flavor (data(i)%flv))
       call state_iterator_init (it_hel, pol%state)
       do while (state_iterator_is_valid (it_hel))
          qn_hel = state_iterator_get_quantum_numbers (it_hel)
          qn = qn_hel(1) .merge. qn_fc(1)
          call interaction_add_state (int, (/ qn, qn, qn_photon /))
          call state_iterator_advance (it_hel)
       end do
       call polarization_final (pol)
    end do
    call interaction_freeze (int)
  end subroutine interaction_init_ewa
    
  elemental subroutine strfun (f, x, xb, r, E, data)
    real(default), intent(out) :: f, x, xb
    real(default), intent(in) :: r, E
    type(ewa_data_t), intent(in) :: data
    real(default) :: rb, lx0, lx1, lx, d, den
    real(default) :: c1, c2, pt2
    real(default) :: fm, fp, fL, fsum
    !!! Here I assume we always do a mapping
    lx0 = log (data%x_0)
    lx1 = log (data%x_1)
    lx = lx1 * x + lx0 * xb
    x = exp(lx)
    !!! I assume that the original (structure) function value should
    !!! be 1
    f = data%coeff * (lx1 - lx0)
    !!! Still some parts missing, compare old implementation
    !!! What's the equivalent of sqrts here???
    pt2 = min ((data%pt_max)**2, (xb * E / 2)**2)
    c1 = log (1 + pt2 / (xb * (data%mass)**2))
    c2 = 1 / (1 + (xb * (data%mass)**2) / pt2)       
    fm = ((data%cv + data%ca)**2 + ((data%cv - data%ca) * &
            xb)**2) / 2 * (c1 - c2)
    fp = ((data%cv - data%ca)**2 + ((data%cv + data%ca) * &
            xb)**2) / 2 * (c1 - c2)    
    fL = (data%cv**2 + data%ca**2) * 2 * xb * c2
    fsum = fp + fm + fL
    f = f * fsum
    if (fsum /= 0) then
       fp = fp / fsum
       fm = fm / fsum
       fL = fL / fsum
    end if 
    !!!   
    !!! INCOMPLETE
    !!!
    !!! f = 0
    !!! rb = 1 - r
    !!! d = data%log ** 2 &
    !!!      -  4 * (r * lx1 * (data%log - lx1) + rb * lx0 * (data%log - lx0))
    !!! if (d <= 0) then
    !!!    return
    !!! else
    !!!    lx = (data%log - sqrt (d)) / 2
    !!! end if
    !!! x = exp (lx)
    !!! xb = 1 - x
    !!! den = data%log - 2 * lx
    !!! if (den <= 0)  return
    !!! qminsq = max (x ** 2 / xb * data%mass ** 2, data%q_min ** 2)
    !!! qmaxsq = min (4 * E ** 2, data%q_max ** 2)
    !!! if (qminsq < qmaxsq) then
    !!!    f = ((xb + x ** 2 / 2) * (qmaxsq / qminsq) &
    !!!         - (1 - x / 2) ** 2 &
    !!!           * log ((x**2 + qmaxsq / E ** 2) / (x**2 + qminsq / E ** 2)) &
    !!!         - xb ** 2 * data%mass ** 2 / qminsq * (1 - qminsq / qmaxsq)) &
    !!!        * ( data%c1 - data%c0 ) / den
    !!! end if
  end subroutine strfun

  subroutine interaction_apply_ewa (int, r, ewa_data)
    type(interaction_t), intent(inout) :: int
    real(default), dimension(:), intent(in) :: r
    type(ewa_data_t), dimension(:), intent(in) :: ewa_data
    type(vector4_t) :: k
    type(splitting_data_t) :: sd
    real(default), dimension(size(ewa_data)) :: f, x, xb
    k = interaction_get_momentum (int, 1)
    sd = new_splitting_data (k, k**2, ewa_data(1)%mass**2, 0._default)
    call strfun (f, x, xb, r(1), energy (k), ewa_data)
    call interaction_set_flavored_values &
         (int, cmplx (f, kind=default), ewa_data%flv, 2)
    call splitting_set_t_bounds (sd, x(1), xb(1))
    !!! JR: Call does not work loke that
    !!! call splitting_narrow_t_bounds (sd, ewa_data(1)%q_min, ewa_data(1)%q_max)
    select case (size (r))
    case (1)
       call splitting_set_collinear (sd)
    case (3)
       call splitting_sample_t (sd, r(2)) 
       call splitting_sample_phi (sd, r(3))
    case default
       print *, "n_rand = ", size (r)
       call msg_bug (" EWA: number of random numbers must be 1 or 3")
    end select
    call interaction_set_momenta &
         (int, split_momentum (k, sd), outgoing=.true.)
  end subroutine interaction_apply_eWa


end module sf_ewa
